import os
import subprocess

POI_TAGS = ["amenity", "builing", "shop", "office", "tourism"]
OSMOSIS_CMD = "osmosis --read-pbf {0} --node-key keyList={1} --write-xml {2}"


class Extractor(object):
    """
    Use `osmosis` to extract POI nodes from an osm file
    
    """
   
    def __init__(self, tags=POI_TAGS):
        self.tags = tags

    def extract(self, source_pbfs, outdir, clean=False):
        extracted = []
        for source in source_pbfs:
            # get filename without the .pbf extension
            fileroot = source.rsplit('/', 1)[-1].rsplit('.', 1)[0]
            target = os.path.join(outdir, fileroot)
            if self._extract(source, target):
                extracted.append(target)
        return extracted

    def _extract(self, source, target):
        key_list = ",".join(self.tags)
        cmd = OSMOSIS_CMD.format(source, key_list, target)
        try:
            subprocess.call(cmd, shell=True)
            return True
        except Exception as e:
            print e
            return False



import os
import re

import requests

import logging
logger = logging.getLogger(__name__)


US_GEOFABRIK = 'http://download.geofabrik.de/north-america/us/{0}-latest.osm.pbf'


class Downloader(object):
    """
    Downloads OSM extracts from GeoFabrik in pbf format

    """
    
    def build_url(self, state):
        return US_GEOFABRIK.format(state.replace(' ', '-').lower())

    def get_filename(self, url):
        filename = url.rsplit('/', 1)[-1]
        return os.path.join('/tmp', filename)
    
    def download(self, states_file):
        """ For each state in states file build url and download file """
        outfiles = []
        with open(states_file, 'r') as f:
            for state in f.read().splitlines():
                url = self.build_url(state)
                log = "Downloading State < {0} > from < {1} >"
                logger.info(log.format(state, url))
                outfiles.append(self._download(url))
        return outfiles

    def _download(self, url):
        """ Dowload file to /tmp """
        tmp = self.get_filename(url)
        f = open(tmp, 'wb')
        res = requests.get(url, stream=True)
        if res.status_code != 200:
            # failed to download, cleanup and raise exception
            f.close()
            os.remove(tmp)
            error = "{0}\n\nFailed to download < {0} >".format(res.content, url)
            raise IOError(error)
        for block in res.iter_content(1024):
            f.write(block)
        f.close()
        return tmp



import sys
import os
import hashlib
import json

import Geohash
import shapely
from shapely.geometry import shape
from shapely.geometry.point import Point
from rtree import index

from reader import FileReader


RTREE_LOCATION = '/tmp/geotweet-rtree-{0}'
# Reference files to be downloaded from S3
AWS_BUCKET = "https://s3-us-west-2.amazonaws.com/jeffrey.alan.meyers.bucket"
COUNTIES_GEOJSON_URL = os.path.join(AWS_BUCKET, "geotweet/us_counties.json")
# use local files if these environment variables are set with a filepath
try:
    COUNTIES_GEOJSON = os.environ['COUNTIES_GEOJSON_LOCAL']
except KeyError:
    COUNTIES_GEOJSON = COUNTIES_GEOJSON_URL


class SpatialLookup(FileReader):
    """ Create a indexed spatial lookup of a geojson file """
    
    idx = None

    def __init__(self, src=None):
        if src:
            if not self.is_valid_src(src):
                error = "Arg src=< {0} > is invalid."
                error += " Must be existing file or url that starts with 'http'"
                raise ValueError(error.format(src))
            # location of index based on hash on input src name
            location = self.get_location(src)
            if not self._exists(location):
                # index does not create so fetch data and build it
                self._build(src, location)
            else:
                # index already exists
                self.idx = index.Rtree(location)
    
    def get_location(self, src):
        digest = self.digest(src)
        if not digest:
            return None
        return RTREE_LOCATION.format(digest)

    def digest(self, src):
        if not src or type(src) != str:
            return None
        m = hashlib.md5()
        if self.is_url(src):
            m.update(src)
        else:
            m.update(os.path.abspath(src))
        return m.hexdigest()

    def _exists(self, location):
        dat = location + ".dat"
        idx = location + ".idx"
        if os.path.isfile(dat) and os.path.isfile(idx):
            return True
        return False

    def get_object(self, point):
        """ lookup object based on point as [longitude, latitude] """
        # first search bounding boxes
        # idx.intersection method modifies input if it is a list
        try:
            tmp = tuple(point)
        except TypeError:
            return None
        # point must be in the form (minx, miny, maxx, maxy) or (x, y)
        if len(tmp) not in [2, 4]:
            return None
        for bbox_match in self.idx.intersection(tmp, objects=True):
            # check actual geometry
            record = bbox_match.object
            if record['geometry'].intersects(Point(tmp)):
                return record['properties']
        return None
   
    def _build(self, src, location):
        """ Build a RTree index to disk using bounding box of each feature """
        geojson = json.loads(self.read(src))
        def generate():
            for i, feature in enumerate(geojson['features']):
                feature['geometry'] = shape(feature['geometry'])
                yield i, feature['geometry'].bounds, feature
        if geojson:
            # create index, flush to disk which disables access then re-enable access
            self.idx = index.Rtree(location, generate())
            self.idx.close()
            self.idx = index.Rtree(location)


class CachedCountyLookup(SpatialLookup):
    """ Dowload counties geojson, build index and provide lookup functionality """

    geohash_cache = {}

    def __init__(self, src=COUNTIES_GEOJSON):
        super(CachedCountyLookup, self).__init__(src=src)

    def get(self, geohash):
        """ lookup state and county based on geohash of coordinates from tweet """
        if geohash in self.geohash_cache:
            # cache hit on geohash
            return self.geohash_cache[geohash]
        # cache miss on geohash
        state = county = None
        coord = Geohash.decode(geohash)
        prop = self.get_object([float(coord[1]), float(coord[0])])
        if prop:
            state = prop['STATE']
            county = prop['COUNTY']
        self.geohash_cache[geohash] = (state, county)
        return state, county



import os
import json

from lxml.etree import iterparse
import boto3


class GeoJSONLoader(object):

    def load(self, json_file, load_func):
        with open(json_file, 'r') as f:
            data = json.loads(f.read().decode('latin-1').encode('utf-8'))
            for feature in data['features']:
                load_func(feature)
        

class OSMLoader(object):
    """
    Process osm files containing POI nodes

    """

    def load(self, poi_files, load_func):
        for poi in poi_files:
            self._load(poi, load_func)

    def _build_data(self, elem):
        data = dict(elem.attrib)
        data["properties"] = self._get_tags(elem)
        data["geometry"] = {
            "type":"Point",
            "coordinates":[float(data["lon"]), float(data["lat"])]
        }
        data["_id"] = data["id"]
        del data["lat"]
        del data["lon"]
        return data

    def _load(self, poi, load_func):
        for event, elem in iterparse(poi):
            # if a node contains any POI tags grab them and append
            # node and tags as json to output file
            if elem.tag == "node":
                # call load function with data (usually write to database)
                data = self._build_data(elem)
                load_func(data)
            # subtree has been fully processed so clear from memory
            if elem.tag in ["node", "way", "relation"]:
                elem.clear()

    def _get_tags(self, elem):
        tags = {}
        for child in list(elem):
            key = child.attrib.get('k')
            if child.tag == "tag":
                tags[key] = child.attrib.get('v')
        return tags






def osm_runner(args):
    """
    Top level runner to download osm data, extract POI nodes and load into mongodb

    """
    logger.info("Starting OSM POI nodes Mongo DB loading process.")
    logger.info("outdir={0}".format(args.output))
    logger.info("states={0}".format(args.states))
    downloaded = download_osm_data(args.states)
    prepare_output_directory(args.output)
    poi_files = extract_points_of_interest(downloaded, args.output)
    load_osm_mongo(poi_files, uri=args.mongo, db=args.db, collection=args.collection)
    logger.info("Finished loading OSM POI nodes into Mongo DB")


def geojson_runner(args):
    """ Top level runner to load State and County  GeoJSON files into Mongo DB """
    logger.info("Starting States and Counties GeoJSON MongoDB DB loading process.")
    load_geo_mongo(args.file, uri=args.mongo, db=args.db, collection=args.collection)
    logger.info("Finished loading {0} into MongoDB".format(args.file))


def prepare_output_directory(outdir):
    """ Clear output directory and create if does not exist """
    logger.info("Cleaning output directory < {0} >".format(outdir))
    try:
        shutil.rmtree(outdir, ignore_errors=True)
    except Exception as e:
        print e
    try:
        os.makedirs(outdir)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(outdir):
            pass


def download_osm_data(states):
    """
    Download data from Geofabrik

    Returns list of paths to each file that was downloaded
    """
    log = "Start data download from Geofabrik. states={0}"
    logger.info(log.format(states))
    # validate if source file containing state names exists when running in batch mode
    if not os.path.isfile(states):
        raise ValueError("state list file < {0} > does not exist".format(states))
    return Downloader().download(states)
    

def extract_points_of_interest(pbf_extracts, outdir):
    """
    Extract only OSM nodes containing specified tags

    Returns list of files containing POI nodes
    """
    log = "Extract POI nodes from downloaded pfb extracts to < {0} >"
    logger.info(log.format(outdir))
    poi_files = Extractor(tags=DEFAULT_POI_TAGS).extract(pbf_extracts, outdir)
    return poi_files


def load_osm_mongo(poi_files, uri=None, db='osm', collection=None):
    """ Read each osm file and convert each node to json and load into mongo """
    logger.info("Mongo URI: {0}".format(uri))
    logger.info("Mongo DB: {0}".format(db))
    logger.info("Mongo Collection: {0}".format(collection))
    logger.info("POI Files to be loaded < {0} >".format(poi_files))
    mongo = MongoGeo(db=db, collection=collection, uri=uri)
    OSMLoader().load(poi_files, mongo.insert)  


def load_geo_mongo(geojson, uri=None, db=None, collection=None):
    """ Load geojson file into mongodb instance """
    logger.info("Mongo URI: {0}".format(uri))
    logger.info("Mongo DB: {0}".format(db))
    logger.info("Mongo Collection: {0}".format(collection))
    logger.info("Geojson File to be loaded: {0}".format(geojson))
    mongo = MongoGeo(db=db, collection=collection, uri=uri)
    GeoJSONLoader().load(geojson, mongo.insert)  





import os
import re

from reader import FileReader


# Reference files to be downloaded from S3
AWS_BUCKET = "https://s3-us-west-2.amazonaws.com/jeffrey.alan.meyers.bucket"
STOPWORDS_LIST_URL = os.path.join(AWS_BUCKET, "geotweet/stopwords.txt")
# use local files if these environment variables are set with a valid filepath
try:
    STOPWORDS_LIST = os.environ['STOPWORDS_LIST_LOCAL']
except KeyError:
    STOPWORDS_LIST = STOPWORDS_LIST_URL


class WordExtractor(FileReader):
    """
    Extract words from a tweet.

    If a provided `src` keyword param references a local file or remote resource
    containing list of stop words, the will be download and used to exclude
    extracted words
    
    """
    def __init__(self, src=STOPWORDS_LIST):
        self.sub_all = re.compile("""[#.!?,"(){}[\]|]|&amp;""")
        self.sub_ends = re.compile("""^[@\\\/~]*|[\\\/:~]*$""")
        self.stopwords = {}
        if src:
            if not self.is_valid_src(src):
                error = "Arg src=< {0} > is invalid."
                error += " Must be existing file or url that starts with 'http'"
                raise ValueError(error.format(src))
            words = self.read(src)
            for word in words.splitlines():
                    self.stopwords[word] = ""

    def clean_unicode(self, line):
        chars = [char for char in line.lower()]
        only_ascii = lambda char: char if ord(char) < 128 else ''
        return str(''.join(filter(only_ascii, chars)))

    def clean_punctuation(self, word):
        return self.sub_ends.sub('', self.sub_all.sub('', word))

    def run(self, line):
        """
        Extract words from tweet

        1. Remove non-ascii characters
        2. Split line into individual words
        3. Clean up puncuation characters
        
        """
        words = []
        for word in self.clean_unicode(line.lower()).split():
            if word.startswith('http'):
                continue
            cleaned = self.clean_punctuation(word)
            if len(cleaned) > 1 and cleaned not in self.stopwords:
                words.append(cleaned)
        return words



import os
import json

import pymongo
from pymongo import MongoClient

import logging
logger = logging.getLogger(__name__)


MONGODB_URI = os.getenv('GEOTWEET_MONGODB_URI', 'mongodb://127.0.0.1:27017')
DB = "test"
COLLECTION = "test"
PORTLAND = [-122.675983, 45.524764]
DEFAULT_TIMEOUT = 5 * 1000


class Mongo(object):
    """ Base wrapper class to connect to mongo and intialize a collection """
    
    def __init__(self, db=DB, uri=MONGODB_URI, collection=COLLECTION):
        timeout = DEFAULT_TIMEOUT
        args = dict(
            connectTimeoutMS=timeout,
            socketTimeoutMS=timeout,
            serverSelectionTimeoutMS=timeout
        )
        self.client = MongoClient(uri, **args)
        self.db = self.client[db]
        self.collection = self.db[collection]

    def insert(self, data):
        try:
            self.collection.insert_one(data)
        except pymongo.errors.DuplicateKeyError as e:
            logger.warn(str(e))
            logger.warn("Record already exists in database. Skipping")
        except pymongo.errors.WriteError as e:
            logger.warn(str(e))
            logger.warn("Write Error")


class MongoGeo(Mongo):

    def __init__(self, *args, **kwargs):
        super(MongoGeo, self).__init__(*args, **kwargs)
        self.collection.ensure_index([("geometry", "2dsphere")])

    def point(self, coordinates):
        return dict(type="Point",coordinates=coordinates)

    def find(self, query=None, limit=None):
        cmd = self.collection
        if query:
            cmd = cmd.find(query)
        else:
            cmd = cmd.find()
        if limit:
            cmd = cmd.limit(limit)
        return cmd

    def near(self, coordinates, distance=None):
        query = dict(geometry={
            "$geoNear":{
                "$geometry":self.point(coordinates),
            }
        })
        if distance:
            query['geometry']['$geoNear']['$maxDistance'] = distance
        return query

    def intersects(self, coordinates):
        return dict(geometry={
            "$geoIntersects":{
                "$geometry":self.point(coordinates)
            }
        })



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='geotweet',
    version='0.1.9',
    description='Fetch geographic tweets from Twitter Streaming API',
    author='Jeffrey Alan Meyers',
    author_email='jeffrey.alan.meyers@gmail.com',
    url='https://github.com/meyersj/geotweet',
    packages=[
        'geotweet',
        'geotweet.mapreduce.utils',
        'geotweet.twitter',
        'geotweet.mongo',
        'geotweet.geo'
    ],
    scripts=['bin/geotweet', 'bin/geoloader'],
    include_package_data=True,
    install_requires=[
        'setuptools',
        'argparse',
        'boto3',
        'python-twitter',
        'requests',
        'mrjob',
        'pymongo',
        'Geohash',
        'Shapely',
        'Rtree',
        'lxml'
    ]
)



import os
import shutil
import errno

from download import Downloader
from extract import Extractor
from load import OSMLoader, GeoJSONLoader
from ..mongo.mongo import MongoGeo

import logging
logger = logging.getLogger(__name__)


DEFAULT_POI_TAGS = ['amenity', 'builing', 'shop', 'office', 'tourism']


class Geo(object):
    
    def __init__(self, args):
        self.source = args.file
        self.mongo = args.mongo
        self.db = args.db
        self.collection = args.collection

    def run(self):
        """ Top level runner to load State and County  GeoJSON files into Mongo DB """
        logger.info("Starting GeoJSON MongoDB loading process.")
        mongo = dict(uri=self.mongo, db=self.db, collection=self.collection)
        self.load(self.source, **mongo)
        logger.info("Finished loading {0} into MongoDB".format(self.source))

    def load(self, geojson, uri=None, db=None, collection=None):
        """ Load geojson file into mongodb instance """
        logger.info("Mongo URI: {0}".format(uri))
        logger.info("Mongo DB: {0}".format(db))
        logger.info("Mongo Collection: {0}".format(collection))
        logger.info("Geojson File to be loaded: {0}".format(geojson))
        mongo = MongoGeo(db=db, collection=collection, uri=uri)
        GeoJSONLoader().load(geojson, mongo.insert)  


class OSM(object):

    def __init__(self, args):
        self.output = args.output
        self.states = args.states
        self.mongo = args.mongo
        self.db = args.db
        self.collection = args.collection

    def run(self):
        """
        Top level runner to download osm data, extract POI nodes and load into mongodb

        """
        log = "Using states list at: < {0} >"
        logger.info(log.format(self.states))
        logger.info("Starting OSM POI nodes Mongo DB loading process.")
        logger.info("outdir={0}".format(self.output))
        logger.info("states={0}".format(self.states))
        # start pipeline 
        downloaded = self.download(self.states)
        self.prep_output(self.output)
        poi_files = self.extract(downloaded, self.output)
        self.load(poi_files, uri=self.mongo, db=self.db, collection=self.collection)
        logger.info("Finished loading OSM POI nodes into Mongo DB")
    
    def download(self, states):
        """
        Download data from Geofabrik

        Returns list of paths to each file that was downloaded
        """
        log = "Start data download from Geofabrik. states={0}"
        logger.info(log.format(states))
        # validate if source file containing state names exists
        if not os.path.isfile(states):
            raise ValueError("state list file < {0} > does not exist".format(states))
        return Downloader().download(states)

    def prep_output(self, outdir):
        """ Clear output directory and create if does not exist """
        logger.info("Cleaning output directory < {0} >".format(outdir))
        try:
            shutil.rmtree(outdir, ignore_errors=True)
        except Exception as e:
            print e
        try:
            os.makedirs(outdir)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(outdir):
                pass

    def extract(self, pbf_extracts, outdir):
        """
        Extract only OSM nodes containing specified tags

        Returns list of files containing POI nodes
        """
        log = "Extract POI nodes from downloaded pfb extracts to < {0} >"
        logger.info(log.format(outdir))
        poi_files = Extractor(tags=DEFAULT_POI_TAGS).extract(pbf_extracts, outdir)
        return poi_files

    def load(self, poi_files, uri=None, db='osm', collection=None):
        """ Read each osm file and convert each node to json and load into mongo """
        logger.info("Mongo URI: {0}".format(uri))
        logger.info("Mongo DB: {0}".format(db))
        logger.info("Mongo Collection: {0}".format(collection))
        logger.info("POI Files to be loaded < {0} >".format(poi_files))
        mongo = MongoGeo(db=db, collection=collection, uri=uri)
        OSMLoader().load(poi_files, mongo.insert)  



import sys
import os
import urllib2


class FileReader(object):
    """ Read file from the local file system or remote url """

    def is_url(self, src):
        return src.startswith('http')

    def is_valid_src(self, src):
        return os.path.isfile(src) or self.is_url(src)

    def read(self, src):
        """ Download GeoJSON file of US counties from url (S3 bucket) """
        geojson = None
        if not self.is_valid_src(src):
            error = "File < {0} > does not exists or does start with 'http'."
            raise ValueError(error.format(src))
        if not self.is_url(src):
            return open(src, 'r').read().decode('latin-1').encode('utf-8')
        response = urllib2.urlopen(src)
        return response.read().decode('latin-1').encode('utf-8')
